"""String constants used to annotate task constructors."""

# Complexity.
EASY = "easy"
HARD = "hard"

# Observation type.
STATE = "features"
VISION = "vision"

# Reward type.
SPARSE = "sparse"
DENSE = "dense"
